package cz.cvut.ok.kapo.paa.sat.engine;

import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

import cz.cvut.ok.kapo.paa.sat.objects.Configuration;
import cz.cvut.ok.kapo.paa.sat.objects.EConstants;
import cz.cvut.ok.kapo.paa.sat.objects.Formula;
import cz.cvut.ok.kapo.paa.sat.objects.SharedVars;
import cz.cvut.ok.kapo.paa.sat.objects.State;

/**
 * Provadi vlastni vypocet. Je volana z vlakna SAThread. Jeji vypocet skonci pouze
 * zavolanim stop() nad vlaknem.
 * @author karel
 *
 */
public class SAEngine {
	public Logger Log = Logger.getLogger(EConstants.APP_LOG);
	protected Configuration conf;
	protected Formula formula;
	protected State best;			//nejlepsi libovolny stav
	protected State actual;			//aktualni stav
	protected State bestSolution;	//zatim nejlepsi pripustny stav
	protected Random rndg;
	protected int iters,accepted,refused;
	protected double fraction;
	
	/**
	 * Inicializuje Engine.
	 * @param form formule, ktera se bude pouzivat.
	 */
	public SAEngine(Configuration conf) {
		super();
		this.conf = conf;
		this.accepted=1; //musi byt pro prvni volani getFraction();
		this.refused=0;
		this.formula = conf.getFormula();
		this.rndg = new Random(System.nanoTime());
		this.actual = this.generateStartingState(1000,true);
		this.best = this.actual;
		this.bestSolution = null;
		this.fraction = 0.0;
	}
	
	/**
	 * Zjisti, zda je dany stav i resenim (Tj. predmety se vlezou do batohu).
	 * @param state seznam predmetu v batohu
	 * @return true, pokud neni batoh preplnen.
	 */
	public boolean isSolution(State state){
		return this.formula.getFormulaValue(state);
	}
	
	/**
	 * Zjisti, zda dany stav ma lepsi hodnotu, nez aktualni.
	 * @return true, pokud predany stav je lepsi.
	 */
	public boolean isBetter(State state){
		return this.getCost(state) > this.getCost(this.actual);
	}
	
	/**
	 * Zjisti, zda dany stav ma lepsi hodnotu, nez nejlepsi.
	 * @return true, pokud predany stav je lepsi.
	 */
	public boolean isBest(State state){
		return this.getCost(state) > this.getCost(this.best);
	}
	
	/**
	 * Vrati prozatim nejlepsi vysledek. Pozor! nemusi to byt reseni!.
	 */
	public State getBest(){
		return this.best;
	}
	
	/**
	 * Zaradi aktualni vysledky do vystupnich 
	 * statistickych front.
	 */
	public void fillStats() {
		synchronized (SharedVars.lock) {
			SharedVars.bestSolutionsStat.add(normalize(this.bestSolution));
			SharedVars.bestWorkingsStat.add(normalize(this.best));
			SharedVars.workingsStat.add( normalize(this.actual));
		}
	}
	
	private double getCost(State state){
		return this.formula.getCost(state);
	}
	
	protected Integer normalize(State state){
		return new Integer((int)(this.getCost(state)));
	}
	
	/**
	 * Vygeneruje 10 stavu, ktere se lisi od puvodniho ve 2 nahodnem bitu, a vybere
	 * z nich ten, ktery ma nejvice splnenych klausuli.
	 * @param old stary obsah batohu
	 * @return novy stav batohu
	 */
	protected State generateNewNeighbour(State old) {
		State best=old.clone();
		int idx = (Math.abs(this.rndg.nextInt()) % best.getVarsCount()) + 1;
		best.setValue(idx, !best.getValue(idx));
		
		/*
		if (this.fraction >= EConstants.mediumFraction) {
			idx = (Math.abs(this.rndg.nextInt()) % best.getVarsCount()) + 1;
			best.setValue(idx, !best.getValue(idx));
			if (this.fraction >= EConstants.highFraction) {
				idx = (Math.abs(this.rndg.nextInt()) % best.getVarsCount()) + 1;
				best.setValue(idx, !best.getValue(idx));
			}
		}*/
	
		/*for (int i = 0; i < 2; i++) {
			State tmp = old.clone();
			idx = Math.abs(this.rndg.nextInt()) % tmp.getVarsCount() + 1;
			tmp.setValue(idx, !tmp.getValue(idx));
			if (this.formula.getRightClausulesCount(tmp) > this.formula.getRightClausulesCount(best))
				best = tmp;
		}*/
		
		return best;
	}

	/**
	 * Generuje stavy pomoci @see generateNew() dokud neni splnena nektera z podminek
	 * pro SA algoritmus.
	 * @param old puvodni stave
	 * @return novy stav
	 */
	protected State tryState(State old, double temperature) {

		State tmp = generateNewNeighbour(old);
		if (this.getCost(tmp) > this.getCost(old)){
			this.accepted++;
			return tmp;
		}
		
		double deltaCost = Math.abs(this.getCost(tmp) - this.getCost(old));
		double x = this.rndg.nextDouble();
		double value = Math.exp(-deltaCost / temperature);
		if (x < value){
			this.accepted++;
			return tmp;
		}
		
		this.refused++;
		return old;
	}
	
	/**
	 * Spusti cast vypoctu. Aktualni stav se uchovava v instancnich promennych.
	 * Vypocet pri dalsim spusteni pokracuje tam, kde minule prestal.
	 * 
	 * @return true, pokud bylo nalezeno optimalni reseni
	 */
	public boolean runPart(double temperature, int N){
		this.iters = 0; //vynulujeme iterace
		this.fraction = getFraction();	//nastavime uspesnost minuleho vypoctu
		this.accepted=this.refused=0;//vynulujeme merice
		this.actual = this.best;
		
		while (!equilibrium(N)){ //vlastni podminka pro ukonceni iteraci na jedne teplote
			this.actual = tryState(this.actual,temperature);
			if (isBest(this.actual)){
				this.best = this.actual.clone();
			}
			if (isSolution(this.actual)){
				if (this.getCost(this.actual) > this.getCost(this.bestSolution)) {
					addSolution(this.actual.clone());
					this.bestSolution = this.actual.clone();
				}
			}
		}

		//fillStats();
		return false;
	}

	/**
	 * Prida do gloalni fronty reseni predane reseni.
	 * @param solution
	 */
	private void addSolution(State solution) {
		synchronized (SharedVars.lock) {
			SharedVars.solutions.add(solution);
		}
	}
	
	/**
	 * urcuje, kdy uz je vhodne prestat iterovat. Zatim je nastavena 
	 * jen jedna podminka: pocet iteraci.
	 * @param N
	 * @return
	 */
	protected boolean equilibrium(int N){
		this.iters++;
		if (this.iters>N)
			return true;
		return false;
	}

	/**
	 * Test, zda doslo ke zmrazeni (malo prijatych zmen - celkove).
	 */
	public boolean frozen(){
		double fraction = getFraction();
		this.Log.log(Level.INFO, "accepted: "+ (int)(fraction*100) + "%");
		if (this.accepted + this.refused < 100)
			return false;
		
		if (fraction < EConstants.frozen)
			return true;
		
		return false;
	}

	/**
	 * Vrati pomer (double)this.accepted / (this.accepted+this.refused)
	 */
	public double getFraction(){
		return (double)this.accepted / (this.accepted+this.refused);
	}
	
	/**
	 * Metodou puleni intervalu najde startovaci teplotu.
	 * Jako pocatecni teplotu bere takovou, pri ktere dosahne tento pomer:
	 * prijatych / (prijatych+neprijatych) pomeru endingRatio s presnosti
	 * 0.05 .
	 * @param startingDelta pocatecni "prihazovani" k teplote
	 * @param endingRatio cilovy pomer prijatych stavu
	 * @return teplota
	 */
	public double findStartingTemperature(double startingDelta, double endingRatio){
		//budeme generovat stavy tak, aby byl pomer
		//lepsich/horsim 1:1.
		//pritom musime menit teplotu:
		// vice lepsich: zvysovat
		// vice horsich: snizovat
		double temperature=0.01;	//pocatecni teplota
		double delta=Math.abs(startingDelta); //potrebujeme zacit zvysovat teplotu
		double ratio=1.0;
		
		while( Math.abs(ratio-endingRatio)>0.02 ){ //opakuj, dokud se nepriblizime dostatecne k pomeru endingRatio
			temperature += delta;
			ratio = testTemperature(temperature); //otestujeme aktualni teplotu
			this.Log.log(Level.INFO, "ratio: "+ratio+" temp: "+temperature+" delta: "+delta);
			if (ratio > endingRatio){
				if (delta>0.0)
					delta *= -0.5;  //moc teplo, musime ochladit
			}
			else {
				if (delta<0.0)
					delta *= -0.5; //moc zima, musime ohrat
			}
		}
		this.accepted=1;
		this.refused=0;
		return temperature;
	}
	
	/**
	 * Provede 100 vygenerovani sousedu a vrati pomer 
	 * prijatych/(prijatych + neprijatych)
	 * stavu.
	 * @param temp - teplota na ktere provest test.
	 * @return
	 */
	protected double testTemperature(double temp){
		this.accepted=this.refused=0;
		State tmp = this.generateRandomState();
		for (int i=0;i<250; i++)
			tmp=this.tryState(tmp,temp);
		return this.getFraction();
	}
	
	/**
	 * Vygeneruje pocatecni stav. Generuje nahodne ohodnoceni promennych.
	 * Snazi se vratit platny stav (tj. isSolution() = true). Pokud se mu to 
	 * nepovede to <b>tries</b> pokusu, bud vrati null (acceptNoSol==false)
	 * nebo posledni vygenerovany stav (acceptNoSol==true).
	 * @param tries - pocet pokusu na nalezeni pripustneho stavu.
	 * @param acceptNoSol - zda ma vratit i stav, ktery neni resenim.
	 * @return stav/null podle parametru.
	 */
	public State generateStartingState(int tries, boolean acceptNoSol){
		/*State st = new State(this.conf.getNumVars());
		for (int i=1; i<= this.conf.getNumVars(); i++)
			if (i%3 == 0)
				st.setValue(i, !st.getValue(i));
		return st;*/
		State tmp=null;
		for (int i=0; i<tries; i++){
			tmp = generateRandomState();
			if (this.isSolution(tmp))
				return tmp;
		}

		if (acceptNoSol)
			return tmp;
		return null; 
	}
	
	/**
	 * Vygeneruje nahodny stav a vrati ho.
	 * @return
	 */
	public State generateRandomState(){
		State tmp = new State(this.conf.getNumVars());
		for (int i=1; i<=this.conf.getNumVars(); i++)
			tmp.setValue(i, this.rndg.nextBoolean());
		return tmp;
	}
}
